@@ -1273,7 +1273,16 @@ public abstract class XnioWorker extends AbstractExecutorService implements Conf
         public Thread newThread(final Runnable r) {
             return doPrivileged(new PrivilegedAction<Thread>() {
                 public Thread run() {
-                    final Thread taskThread = new Thread(threadGroup, r, name + " task-" + getNextSeq(), stackSize);
+                    final Thread taskThread = new Thread(threadGroup, new Runnable() {
+                        @Override
+                        public void run() {
+                            try {
+                                r.run();
+                            } finally {
+                                xnio.handleThreadExit();
+                            }
+                        }
+                    }, name + " task-" + getNextSeq(), stackSize);
                     // Mark the thread as daemon if the Options.THREAD_DAEMON has been set
                     if (markThreadAsDaemon) {
                         taskThread.setDaemon(true);
