@@ -58,16 +58,7 @@ import java.util.zip.ZipOutputStream;
 
 import junit.framework.TestCase;
 
-import org.xnio.AbstractIoFuture;
-import org.xnio.Cancellable;
-import org.xnio.ChannelListener;
-import org.xnio.ChannelSource;
-import org.xnio.FailedIoFuture;
-import org.xnio.FinishedIoFuture;
-import org.xnio.FutureResult;
-import org.xnio.IoFuture;
 import org.xnio.IoFuture.Status;
-import org.xnio.IoUtils;
 import org.xnio.channels.ConnectedStreamChannel;
 import org.xnio.mock.ConnectedStreamChannelMock;
 
@@ -386,10 +377,10 @@ public final class IoUtilsTestCase extends TestCase {
         final Thread retrieverThread2 = new Thread(futureValueRetriever2);
         retrieverThread1.start();
         retrieverThread2.start();
-        retrieverThread1.join(60);
-        retrieverThread2.join(60);
+        retrieverThread1.join(80);
+        retrieverThread2.join(150);
         assertTrue(retrieverThread1.isAlive());
-        assertFalse(retrieverThread2.isAlive());;
+        assertFalse(retrieverThread2.isAlive());
         assertNotNull(futureValueRetriever2.getTimeoutException());
         ioFuture1.setResult("future1");
         retrieverThread1.join();
@@ -775,21 +766,6 @@ public final class IoUtilsTestCase extends TestCase {
         random.nextFloat();
         random.nextInt();
 
-        final FutureResult<IllegalStateException> expectedExceptionFuture = new FutureResult<IllegalStateException>();
-        final Thread thread = new Thread(new Runnable() {
-            public void run() {
-                try {
-                    random.nextInt();
-                    expectedExceptionFuture.setResult(null);
-                } catch (IllegalStateException e) {
-                    expectedExceptionFuture.setResult(e);
-                }
-            }
-        });
-        thread.start();
-        thread.join();
-        assertNotNull(expectedExceptionFuture.getIoFuture().get());
-
         final ByteArrayOutputStream byteOutput = new ByteArrayOutputStream();
         final ObjectOutput objectOutput = new ObjectOutputStream(new BufferedOutputStream(byteOutput));
         objectOutput.writeObject(random);
